open Defs
open Printf


type value = 
    VInteger of int
  | VReal of float
  | VBoolean of bool
  | VUndef


type 'a codevalue =
    CodeI of ('a,int) code
  | CodeR of ('a,float) code
  | CodeB of ('a,bool) code

exception RealConversion

let mkReal x = match x with
  CodeI i -> .< float_of_int .~i >.
| CodeR r -> r
| _ -> raise RealConversion 

let unCI (CodeI i) = i
let unCR (CodeR r) = r
let unCB (CodeB b) = b


let assign_value ((s:string),(v:'a codevalue)) (memI,memR,memB)
    = match Hashtbl.find symtab s with
	(TInteger,adr) -> .< (.~memI).(adr) <- .~(unCI v) >.
      | (TReal   ,adr) -> .< (.~memR).(adr) <- .~(mkReal v) >.
      | (TBoolean,adr) -> .< (.~memB).(adr) <- .~(unCB v) >.

let lookup_mem ((inf:info),(s:string)) (memI,memR,memB) = 
  let adr = lookup_address (inf,s) in
  match inf.typ with
    TInteger -> CodeI .< (.~memI).(adr) >.
  | TReal    -> CodeR .< (.~memR).(adr) >.  
  | TBoolean -> CodeB .< (.~memB).(adr) >.  

 

let rec codeGenProgram (Program (_,_,sequence)) mem = codeGenSeq sequence mem
and codeGenSeq xs mem = match xs with
  [] -> .< () >.
| (y::ys) -> .< begin .~(codeGenInfoStmt y mem); 
                      .~(codeGenSeq ys mem) end >.
and codeGenInfoStmt (inf,s) mem = match s with
  Assign (s,e) -> assign_value (s, codeGenInfoExpr e mem) mem
| Read s -> 
    begin
      match lookup_type (inf,s) with
	TInteger -> 
	  .< begin
              Printf.printf "integer     %s <- " s;
              let i = read_int() in
              .~(assign_value (s, CodeI .<i>.) mem)
             end >.
      | TReal -> 
	  .< begin
              Printf.printf "real number %s <- " s;
              let r = read_float() in          
              .~(assign_value (s, CodeR .<r>.) mem) 
             end >.
      | _ -> .< Printf.printf "input type unknown of %s" s >.
    end
| Write e -> 
    begin
      match (gI e).typ with
        TInteger -> .< Printf.printf "integer       -> %d\n" .~(unCI (codeGenInfoExpr e mem)) >.
      | TReal -> .< Printf.printf "real number   -> %f\n" .~(mkReal (codeGenInfoExpr e mem)) >.
      | TBoolean -> .< if .~(unCB (codeGenInfoExpr e mem)) 
                      then Printf.printf "boolean value   -> true\n"
                      else Printf.printf "boolean value   -> false\n" >.
    end
| Sequence xs -> codeGenSeq xs mem
| IfThenElse (cond,cthen,celse) ->
    let codeCond = unCB (codeGenInfoExpr cond mem) in
    let codeThen = codeGenInfoStmt cthen mem in
    let codeElse = 
      match celse with
	None -> .< () >.
      | Some stmt ->  codeGenInfoStmt stmt mem in
    .< if .~codeCond then .~codeThen else .~codeElse >.  
| While (cond,body) ->
    .< while .~(unCB (codeGenInfoExpr cond mem)) do 
         .~(codeGenInfoStmt body mem) 
       done >.
| For (var,estart,eend,vstep,body) ->
    let cstart = unCI (codeGenInfoExpr estart mem) in
    let cend = unCI (codeGenInfoExpr eend mem) in
    let cbody i = .< begin .~(assign_value (var,CodeI i) mem); 
                           .~(codeGenInfoStmt body mem) end >. in
    if vstep>0 
    then .< for i = .~cstart to .~cend do .~(cbody .<i>.) done >.
    else .< for i = .~cstart downto .~cend do .~(cbody .<i>.) done >.
and codeGenInfoExpr (inf,e) mem = match e with
  CInteger i -> CodeI .<i>.
| CReal r -> CodeR .<r>.
| Time -> CodeR .< Unix.gettimeofday() >.
| CBoolean b -> CodeB .<b>.
| Var s -> lookup_mem (inf,s) mem
| Par e -> codeGenInfoExpr e mem
| Op (op,xs) -> codeGenOp (gE op) xs mem
and codeGenOp op xs mem = let ys = List.map (fun x -> codeGenInfoExpr x mem) xs in
                    applyOp op ys
and applyOp op xs = match (op,xs) with
   (OAdd, [CodeI a; CodeI b]) -> CodeI .< .~a + .~b >.
|  (OAdd, [a;b]) -> CodeR .< .~(mkReal a) +. .~(mkReal b) >.
|  (OSub, [CodeI a; CodeI b]) -> CodeI .< .~a - .~b >.
|  (OSub, [a;b]) -> CodeR .< .~(mkReal a) -. .~(mkReal b) >.
|  (OMul, [CodeI a; CodeI b]) -> CodeI .< .~a * .~b >.
|  (OMul, [a;b]) -> CodeR .< .~(mkReal a) *. .~(mkReal b) >.
|  (OFrac, [a;b]) -> CodeR .< .~(mkReal a) /. .~(mkReal b) >.
|  (ODiv, [CodeI a; CodeI b]) -> CodeI .< .~a / .~b >.
|  (OMod, [CodeI a; CodeI b]) -> CodeI .< .~a mod .~b >.
|  (OPos, [CodeI a]) -> CodeI a
|  (OPos, [CodeR a]) -> CodeR a
|  (ONeg, [CodeI a]) -> CodeI .< -  .~a >.
|  (ONeg, [CodeR a]) -> CodeR .< -. .~a >.
|  (ONot, [CodeB a]) -> CodeB .< not .~a >.
|  (OAnd, [CodeB a; CodeB b]) -> CodeB .< .~a && .~b >.
|  (OOr, [CodeB a; CodeB b]) -> CodeB .< .~a || .~b >.
|  (OEq, [CodeI a; CodeI b]) -> CodeB .< .~a = .~b >.
|  (OEq, [CodeI a; CodeR b]) -> CodeB .< float_of_int .~a = .~b >.
|  (OEq, [CodeR a; CodeI b])  -> CodeB .< .~a = float_of_int .~b >.
|  (OEq, [CodeR a; CodeR b]) -> CodeB .< .~a = .~b >.
|  (OEq, [CodeB a; CodeB b]) -> CodeB .< .~a = .~b >.
|  (OLess, [CodeI a; CodeI b]) -> CodeB .< .~a < .~b >.
|  (OLess, [CodeI a; CodeR b]) -> CodeB .< float_of_int .~a < .~b >.
|  (OLess, [CodeR a; CodeI b]) -> CodeB .< .~a < float_of_int .~b >.
|  (OLess, [CodeR a; CodeR b]) -> CodeB .< .~a < .~b >.
|  (OLess, [CodeB a; CodeB b]) -> CodeB .< .~a < .~b >.
|  (ONeq, [x;y]) -> CodeB .< not .~(unCB (applyOp OEq [x;y])) >.
|  (OLeq, [x;y]) 
   -> CodeB .<    .~(unCB (applyOp OEq   [x;y]))
               || .~(unCB (applyOp OLess [x;y])) >.
|  (OGeq, [x;y]) -> CodeB .< not .~(unCB (applyOp OLess [x;y])) >.
|  (OGrea, [x;y]) -> CodeB .< not .~(unCB (applyOp OLeq  [x;y])) >.

let codeGen prg = 
 .<  let memI = Array.make !curr_address_Integer 0 in
     let memR = Array.make !curr_address_Real 0.0 in
     let memB = Array.make !curr_address_Boolean false in
     .~(codeGenProgram prg (.<memI>.,.<memR>.,.<memB>.)) >.

