open Defs
open Printf


type value = 
    VInteger of int
  | VReal of float
  | VBoolean of bool
  | VUndef

let unVI (VInteger i) = i
let unVR (VReal r) = r
let unVB (VBoolean b) = b

let mkReal x = match x with
  VInteger i -> float_of_int i
| VReal r -> r
| _ -> Printf.printf "error in conversion to real"; exit(0)


let assign_value ((s:string),(v:value)) (memI,memR,memB)
    = match Hashtbl.find symtab s with
	(TInteger,adr) -> memI.(adr) <- (unVI v) 
      | (TReal   ,adr) -> memR.(adr) <- (mkReal v)
      | (TBoolean,adr) -> memB.(adr) <- (unVB v)

let lookup_mem ((inf:info),(s:string)) (memI,memR,memB) = 
  let adr = lookup_address (inf,s) in
  match inf.typ with
    TInteger -> VInteger (memI.(adr))
  | TReal    -> VReal    (memR.(adr))  
  | TBoolean -> VBoolean (memB.(adr))  

let mkReal x = match x with
  VInteger i -> float_of_int i
| VReal r -> r
| _ -> Printf.printf "error in conversion to real"; exit(0)

let rec interProgram (Program (_,_,sequence)) mem = interSeq sequence mem
and interSeq xs mem = List.iter (fun x -> interInfoStmt x mem) xs
and interInfoStmt (inf,s) mem = match s with
  Assign (s,e) -> let v = interInfoExpr e mem in
                  let t = (gI e).typ in
                  assign_value (s, if t=TReal then VReal (mkReal v) else v) mem
| Read s -> 
    begin
      match lookup_type (inf,s) with
	TInteger -> 
	  Printf.printf "integer     %s <- " s;
          let i = read_int() in
          assign_value (s, VInteger i) mem
      | TReal -> 
	  Printf.printf "real number %s <- " s;
          flush stdout;
          let r = read_float() in          
          assign_value (s, VReal r) mem
      | _ -> Printf.printf "input type unknown of %s" s
    end
| Write e -> 
    begin
      match interInfoExpr e mem with
        VInteger i -> Printf.printf "integer       -> %d\n" i
      | VReal r ->    Printf.printf "real number   -> %f\n" r
      | VBoolean true  ->    Printf.printf "boolean value   -> true\n"
      | VBoolean false ->    Printf.printf "boolean value   -> false\n"
    end
| Sequence xs -> interSeq xs mem
| IfThenElse (cond,cthen,celse) ->
    begin
      match (interInfoExpr cond mem, celse) with
	(VBoolean true,_) -> interInfoStmt cthen mem
      | (VBoolean false, Some stmt) -> interInfoStmt stmt mem
      | _ -> ()
    end 
| While (cond,body) ->
    begin
      let c = ref false in
      let VBoolean c' = interInfoExpr cond mem in
      c := c';
      while !c do 
	interInfoStmt body mem;
	let VBoolean c' = interInfoExpr cond mem in
        c := c'
      done
    end
| For (var,estart,eend,vstep,body) ->
    let vcurr = ref (match interInfoExpr estart mem with VInteger i -> i) in
    let VInteger vend = interInfoExpr eend mem in
    while (if vstep>0 then !vcurr<=vend else !vcurr>=vend) do
      assign_value(var,VInteger !vcurr) mem;
      interInfoStmt body mem;
      if vstep>0 then incr vcurr else decr vcurr
    done
and interInfoExpr (inf,e) mem = match e with
  CInteger i -> VInteger i
| CReal r -> VReal r
| Time  -> VReal (Unix.gettimeofday())
| CBoolean b -> VBoolean b
| Var s -> lookup_mem (inf,s) mem
| Par e -> interInfoExpr e mem
| Op (op,xs) -> interOp (gE op) xs mem
and interOp op xs mem = let ys = List.map (fun x -> interInfoExpr x mem) xs in
                        applyOp op ys
and unB (VBoolean b) = b
and applyOp op xs = match (op,xs) with
   (OAdd, [VInteger a; VInteger b]) -> VInteger (a+b)
|  (OAdd, [a;b]) -> VReal (mkReal a +. mkReal b)
|  (OSub, [VInteger a; VInteger b]) -> VInteger (a-b)
|  (OSub, [a;b]) -> VReal (mkReal a -. mkReal b)
|  (OMul, [VInteger a; VInteger b]) -> VInteger (a*b)
|  (OMul, [a;b]) -> VReal (mkReal a *. mkReal b)
|  (OFrac, [a;b]) -> VReal (mkReal a /. mkReal b)
|  (ODiv, [VInteger a; VInteger b]) -> VInteger (a/b)
|  (OMod, [VInteger a; VInteger b]) -> VInteger (a mod b)
|  (OPos, [VInteger a]) -> VInteger a
|  (OPos, [VReal a]) -> VReal a
|  (ONeg, [VInteger a]) -> VInteger (-a)
|  (ONeg, [VReal a]) -> VReal (-. a)
|  (ONot, [VBoolean a]) -> VBoolean (not a)
|  (OAnd, [VBoolean a; VBoolean b]) -> VBoolean (a && b)
|  (OOr, [VBoolean a; VBoolean b]) -> VBoolean (a || b)
|  (OEq, [VInteger a; VInteger b]) -> VBoolean (a=b)
|  (OEq, [VInteger a; VReal b]) -> VBoolean (float_of_int a = b)
|  (OEq, [VReal a; VInteger b])  -> VBoolean (a = float_of_int b)
|  (OEq, [VReal a; VReal b]) -> VBoolean (a=b)
|  (OEq, [VBoolean a; VBoolean b]) -> VBoolean (a=b)
|  (OLess, [VInteger a; VInteger b]) -> VBoolean (a<b)
|  (OLess, [VInteger a; VReal b]) -> VBoolean (float_of_int a < b)
|  (OLess, [VReal a; VInteger b]) -> VBoolean (a < float_of_int b)
|  (OLess, [VReal a; VReal b]) -> VBoolean (a<b)
|  (OLess, [VBoolean a; VBoolean b]) -> VBoolean (a<b)
|  (ONeq, [x;y]) -> VBoolean (not (unB (applyOp OEq [x;y])))
|  (OLeq, [x;y]) 
   -> VBoolean (     (unB (applyOp OEq   [x;y]))
                  || (unB (applyOp OLess [x;y])) )
|  (OGeq, [x;y]) -> VBoolean (not (unB (applyOp OLess [x;y])))
|  (OGrea, [x;y]) -> VBoolean (not (unB (applyOp OLeq  [x;y])))


let interpret prg = 
   let memI = Array.make !curr_address_Integer 0 in
   let memR = Array.make !curr_address_Real 0.0 in
   let memB = Array.make !curr_address_Boolean false in
   interProgram prg (memI,memR,memB)

